#!/bin/sh
# The next line is executed by /bin/sh, but not Tcl \
    exec /usr/local/bin/wish8.3 $0 ${1+"$@"}

# if you are installing by hand
# replace /usr/local/bin/wish8.3
# on line 3 with the full path to your wish executable, verison 8.3 or higher.

global startupdir
set startupdir ""
set scriptname [info script]
# on unix, chase symbolic links until we get to the real program
if {"$tcl_platform(platform)" == "unix"} {
    while {[catch {set scriptname [file readlink $scriptname]}] == 0} { }
}
set startupdir [file dirname $scriptname]
if {"$startupdir" == "."} {
     set startupdir [pwd]
}
lappend auto_path $startupdir [file dirname $startupdir]

package require mkWidgets
package require msTclObjs

option add *Dialog.msg.wrapLength 700
option add *Dialog.msg.font {helvetica 12 bold}

global treeObjId
set treeObjId 0

global title
set title "MapServer Editor"

global mapobj	
set mapobj  ""

global mapfile	
set mapfile ""

global cutbuffer
set cutbuffer ""

global tree	
global edit	

global maptypes
set maptypes {
    {{Map Files}        {.map}       }
    {{Map Backup Files} {.map.bak}   }
    {{All Files}        *            }
}

global dirty
set dirty 0

global showattrnodes
set showattrnodes 0

global editobj edittreeobj 
set editobj ""
set edittreeobj ""

global editattrs postedit preedit
array set editattrs {}
array set postedit {}
array set preedit {}


# images for tree node icons

global node_img
array set node_img {
_default 
{R0lGODlhEAAQAPcAAAAAAAAAhAAA/wD///8AAP//////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
/////////////////////////////////yH5BAEAAAQALAAAAAAQABAAAAhYAAkIHEiwYEEB
CBMKMDhQwICHDwMwFOhQocSJFRNeZCgggMePEwkCCGlwJEkAKFOmZGiyYEsCChPCjLnQIcQB
Em1CzDnAIsyeGn/6zIgw50eQHY9uJEkwIAA7}

_attr
{R0lGODdhEAAQAKEAAP///wAA/wD//wAAACwAAAAAEAAQAAACIISPqcvtD1mYQYhGqV05X9Vp
XFiN4Qd2m6ShiwVHslMAADs=}

class
{R0lGODdhEAAQAMIAAPj8APj8uAAA+KjU+PgA+PjY+AAAAMjMyCwAAAAAEAAQAAADNAi63BAw
ykmnuNVeHLb/3iCOZGmeqEis6bi+5ivPcmHfeH4bvK7zQN8OSCwCD8ikcslsKhMAOw==}

feature
{R0lGODdhEAAQAKEAAP///8zMzAAAAAAAACwAAAAAEAAQAAACKoSPqcsaIlKQMKaK4Xp5six4
GygqVcCVCOSli3Zwr3WwL0U3BqzvuQ4rAAA7}

join
{R0lGODdhEAAQAOMAAPj8+AAAAPjAyAD8+OAAAPgAAAAA+AAA4MjMyPgA+MB8+OCI+AAAAAAA
AAAAAAAAACwAAAAAEAAQAAAEYBDISSsIIotAw/hDEBBZwU2B8Rki4RKnFByHcbRvfNE27uoz
222U69BoPhhCgpgdcYVoANFMGK7DQiKhTYgSCoVV1AxwudqwdzopJxZwMNvShMPnlot9obME
9n0Vf3Z9EQA7}

label
{R0lGODdhEAAQAKEAAP///8zMzAAAAAAAACwAAAAAEAAQAAACLYSPqRexGoR4SQpHQbQZzJ1F
hpRN4/WgBghZrlm9XCLG2Aojs1bTEuaKBTuJAgA7}

layer
{R0lGODdhEAAQAMIAAP////8AAP/JtQDMAKL/iQAA/7Lh/wAAACwAAAAAEAAQAAADRAiq0f5h
MUGrlSDYLdT4DVd536CJHfARRCiSbAs5yyrPQaHbcc/qu4HPVzAYS8ikzmhIOpfMqNQAgE6l
VaB2q1Nwv4UEADs=}

legend
{R0lGODdhEAAQAMIAAP////8AAAAAAAD/AAAA/wAAAAAAAAAAACwAAAAAEAAQAAADMgi63Bow
BrdkVALkDCzEWth5VGme1KCug8mu4Ka8aizSLaoDRO8Tpp/PxhH2iAojkJIAADs=}

map
{R0lGODdhEAAQAKEAAP///wD//wAA/wD/ACwAAAAAEAAQAAACNYQRohvnMaIcork1pTo4Z9R5
UbUIolYqJ5Wqo9qGcNzR6dsmcg6UWcpxTRiO3q9iKZpeSUABADs=}

points
{R0lGODdhEAAQAKEAAPj8+AAAAAD8+AAA+CwAAAAAEAAQAAACLISPqavhjJyM70gnzA32jiFs
3QQMohFW2oUODFlChjvLgEDbc6bD0PljyTYFADs=}

projection 
{R0lGODdhEAAQAMIAAPiooAAAAMjMyPj8+AD8+AAA+PgAAAAAACwAAAAAEAAQAAADRwgKETIw
xhWgk1LVuPHQEFFcHhMWaBcZ5iAWg2oYRIta0ExANhwPtMzm9ittXr/g5Hjb6Hidl4wGgowe
GFYL5/lUSSUK2JMAADs=}

query
{R0lGODdhEAAQAKEAAP///8zMzAAAAAAAACwAAAAAEAAQAAACJYSPqRDii5wUELhgGlUv9Ygl
TYVopCGFkKZW3/mS8UnX3kbPQAEAOw==}

querymap
{R0lGODdhEAAQAKEAAAD8AAAAAPj8AAAAACwAAAAAEAAQAAACKoQRp3vJj2CYyjHb5NULVzOF
YVcJpuBhwYmKVMOm5SnDNIm4b7N/0PYoAAA7}

reference
{R0lGODdhEAAQAMIAAP///wAAAMzMzP//AAAA/wAAAAAAAAAAACwAAAAAEAAQAAADRgi6Gx4M
uBaEnapiZy/kwhQMVfcQ5TeQnRQQaOisrODC6bhuL27PJMgNl9HxYKjiTuhAMo1PpzJYlLqW
VeKVOkxuj1rgJgEAOw==}

scalebar
{R0lGODdhEAAQAKEAAMzMzAAAAP///wAAACwAAAAAEAAQAAACMYQRosuCEVBr8Mipor0TGc1N
lZVg3hGV4jeq1IhiAkjKGu0ycIzhkeyBaGSzHXG2KQAAOw==}

style
{R0lGODdhEAAQAMIAAAAAAPhooAD8ABAo+LAo+AAAAAAAAAAAACwAAAAAEAAQAAADSQixIA4j
AkJXAFhAiYm6GBYNIcZ4oqSZF+WNzWN1FBQ6Gkij9xOSrtAFlwLWZjhY56PTcISMkgra+pGI
s6ASG9KWiKWgMLayBRMAOw==}

symbol
{R0lGODdhEAAQAKEAAP///wAAAP8A/wAAACwAAAAAEAAQAAACMUQCaai3nuA7NNZ5BA1CU595
XRdu44lu6sq2KwqDXHya9OiENPTF1wP7WUrCBksBKAAAOw==}

web
{R0lGODdhEAAQAIAAAAAAAP///ywAAAAAEAAQAAACKYwPcKgb6V5zc1WW5jU674dQHsRRUWid
Zao2ntWNXKaytT23FcnTWFYAADs=}

} ;# end of array set node_img




proc mkMain {} {
    global tree edit obj_node_uu node_dot_uu title node_img
   
    wm title . $title
    wm protocol . WM_DELETE_WINDOW doExit
    wm geom . 900x600

    menu .menuBar -tearoff 0

    .menuBar add cascade -menu .menuBar.file -label "File" -underline 0
    .menuBar add cascade -menu .menuBar.view -label "View" -underline 0
    .menuBar add cascade -menu .menuBar.tool -label "Tools" -underline 0
    .menuBar add cascade -menu .menuBar.help -label "Help" -underline 0

    menu .menuBar.file -tearoff 0
    .menuBar.file add command -label "New"     -underline 0 -command newMap \
	-accelerator Ctrl+N
    bind all <Control-KeyPress-n> +newMap
    .menuBar.file add command -label "Open"    -underline 0 -command openMap \
	-accelerator Ctrl+O
    bind all <Control-KeyPress-o> +openMap
    .menuBar.file add command -label "Save"    -underline 0 -command saveMap \
	-accelerator Ctrl+S
    bind all <Control-KeyPress-s> +saveMap
    .menuBar.file add command -label "Save As" -underline 5 -command saveMapAs \
	-accelerator Ctrl+A
    bind all <Control-KeyPress-a> +saveMapAs
    .menuBar.file add separator
    .menuBar.file add command -label "Quit"    -underline 0 -command doExit \
	-accelerator Ctrl+Q
    bind all <Control-KeyPress-q> +doExit

    menu .menuBar.view -tearoff 0
    .menuBar.view add radiobutton -label "show object attributes" \
		-command redrawTree -variable showattrnodes -value 1 \
		-accelerator Ctrl+W
    bind all <Control-KeyPress-w> "set showattrnodes 1;redrawTree"
    .menuBar.view add radiobutton -label "hide object attributes" \
		-command redrawTree -variable showattrnodes -value 0 \
		-accelerator Ctrl+H
    bind all <Control-KeyPress-h> "+set showattrnodes 0;redrawTree"

    menu .menuBar.tool -tearoff 0
    .menuBar.tool add command -label "Validate mapfile with Mapscript" \
		-command parseWithMS -underline 0 -accelerator Ctrl+V
    bind all <Control-KeyPress-v> +parseWithMS
    .menuBar.tool add separator
    .menuBar.tool add command -label "Start MapServer Explorer with mapfile" \
		-command startMSex -underline 16 -accelerator Ctrl+E
    bind all <Control-KeyPress-x> +startMSex
    .menuBar.tool add command -label "Start MapServer Shape Explorer" \
		-command startMShapex -underline 19 -accelerator Ctrl+P
    bind all <Control-KeyPress-p> +startMShapex

    menu .menuBar.help -tearoff 0
    .menuBar.help add command -label "About" -command loadHelp -underline 0


    . configure -menu .menuBar
    
    pane .top -resize both -orientation vertical -width 3
    frame .top.left  -width 300
    frame .top.right  -width 600
    .top pack first  .top.left  -fill both -expand 1 
    .top pack second .top.right -fill both -expand 1

    frame .top.right.edit -borderwidth 4 -relief sunken -width 400

    treecontrol .top.left.tree -selectmode single -text "(no mapfile opened)" \
		-onselect "checkSelect"
    set tree .top.left.tree
    $tree.text configure -width 32

    # create images for object nodes
    foreach {img_name img_data} [array get node_img] {
	image create photo img_$img_name -data $img_data
    }

    pack .top.left.tree  -fill both -expand 1 -padx 4 -pady 2
    pack .top.right.edit -fill both -expand 1 -padx 4 -pady 2
    pack .top       -side top -fill both -expand 1
    
    label .msg -text ready -anchor w
    pack .msg -side bottom -fill x -anchor w

    
    frame .bottom 
    pack  .bottom -side top -fill x 

    # create a scrolling canvas to hold an inner frame
    # borrowed code from mclennan/harrison, effective tcl/tk programming
    frame .top.right.edit.scrl
    canvas .top.right.edit.scrl.c -yscrollcommand ".top.right.edit.scrl.y set" \
				  -xscrollcommand ".top.right.edit.scrl.x set"
    scrollbar .top.right.edit.scrl.y -orient vertical \
				-command ".top.right.edit.scrl.c yview"
    scrollbar .top.right.edit.scrl.x -orient horizontal \
				-command ".top.right.edit.scrl.c xview"

    pack .top.right.edit.scrl.y -side right -fill y -anchor e
    pack .top.right.edit.scrl.x -side bottom -fill x -anchor s
    pack .top.right.edit.scrl.c -side left -fill both -expand 1 -anchor nw

    pack .top.right.edit.scrl -fill both -expand 1

    frame .top.right.edit.scrl.c.form
    .top.right.edit.scrl.c create window 0 0 -anchor nw \
		-window .top.right.edit.scrl.c.form
    bind .top.right.edit.scrl.c.form <Configure> \
		"scrl_form_resize .top.right.edit.scrl.c"


    set edit .top.right.edit.scrl.c.form
    frame $edit.help
    label $edit.help.title -text "$title\n\nTom Poindexter\n\
            tpoindex@nyx.net\nCopyright 2000\n" -font {times 18 bold italic} 
    pack $edit.help.title -side top -fill x -expand 1 -padx 10 -pady 10
    pack $edit.help -side top -fill x -expand 1
    update 
}

proc loadBanner {} {
    global edit
    pack forget [pack slaves $edit]
    pack $edit.help -side top -fill x -expand 1
}

proc scrl_form_resize {can_win} {
    set bbox [$can_win bbox all]
    set wid [winfo width $can_win.form]
    $can_win configure -width $wid -scrollregion $bbox \
	-yscrollincrement 0.1i -xscrollincrement 0.1i

}

proc getAllMapObjTypes {objtype} {
    lappend subtypes $objtype
    set obj [mstclobj::${objtype}TclObj #auto {}]
    foreach sub [$obj getValidSubObjs] {
	eval lappend subtypes [getAllMapObjTypes $sub]
    }
    obstcl::delete $obj
    return [lsort -unique $subtypes]
}

proc mkEditors {} {
    global edit editattrs preedit postedit
    set ifn {helvetica 10}
    # collect all possible types
    set types_map [getAllMapObjTypes map]
    set types_all $types_map
    # if supporting symbolset objs
    #set types_sym [getAllMapObjTypes symbolset]
    #set types_all [lsort -unique [concat $types_map $types_sym]]
 
    foreach objtype $types_all {
	set obj [mstclobj::${objtype}TclObj #auto {}]
        frame $edit.$objtype
	frame $edit.$objtype.top__bar -relief raised
	label $edit.$objtype.top__bar.attr_type_name \
			-text "Object: $objtype" -anchor w
	pack $edit.$objtype.top__bar.attr_type_name -side left -fill x -anchor w
        button $edit.$objtype.top__bar.val_save -text "validate/store" \
		-command saveEditAttrs -font $ifn
        button $edit.$objtype.top__bar.revert -text "revert" \
		-command revertEdit -font $ifn
	pack $edit.$objtype.top__bar.val_save -padx 6 -side left -anchor w
	pack $edit.$objtype.top__bar.revert   -padx 6 -side left -anchor w
	pack $edit.$objtype.top__bar -side top -fill x -anchor w -pady 6
	foreach {attrname elementnames} [$obj getAttrName] {
	    set af $edit.$objtype.$attrname
	    frame $af -relief raised -borderwidth 2
	    set elementtypes [$obj getAttrType $attrname]
	    set attr_label [lindex $elementnames 0]
	    set subnames [lrange $elementnames 1 end]
	    if {[llength $subnames] == 0} {set subnames [list { }]}
	    label $af.l -text $attr_label -anchor e -width 18 \
			-font {helvetica 12 bold}
	    pack $af.l -side left -anchor w 
	    pack $af -side top    -fill x 
            set i 0
	    foreach subname $subnames {
		set preedit($objtype,$attrname,$i) ""
		set postedit($objtype,$attrname,$i) ""
		if {[string length [string trim $subname]]} {
		    label $af.lab_$i -text $subname -anchor w -font $ifn
		    pack  $af.lab_$i -side left -padx 2 -anchor w
		}
		set subtype [lindex $elementtypes $i]
		switch [lindex $subtype 0] {
		    opt {
			checkbutton $af.ele_$i -font $ifn \
				-onvalue 1 -offvalue "#none" \
				-variable editattrs($attrname,$i)
		    }
		    truth -
		    bool {
                        if {"[lindex $subtype 0]" == "truth"} {
			    set onvalue  true
			    set offvalue false
			} else {
			    set onvalue  on
			    set offvalue off
			}
			frame $af.ele_$i
			radiobutton $af.ele_$i.on  -font $ifn -text $onvalue \
				-value $onvalue \
				-variable editattrs($attrname,$i)
			radiobutton $af.ele_$i.off -font $ifn -text $offvalue \
				-value $offvalue \
				-variable editattrs($attrname,$i)
			radiobutton $af.ele_$i.not -font $ifn \
				-text "default" \
				-value "" \
				-variable editattrs($attrname,$i)
			pack $af.ele_$i.on $af.ele_$i.off \
				$af.ele_$i.not -side left
			set preedit($objtype,$attrname,$i) "
				setAttrToLower $attrname,$i
			"
		    }
		    list {
			set listitems [lindex $subtype 1]
			combobox $af.ele_$i -entries $listitems
			$af.ele_$i configure \
				-textvariable editattrs($attrname,$i)
			if {[set idx [lsearch $listitems \
					$editattrs($attrname,$i)]] >= 0} {
			    $af.ele_$i see $idx
			}
			set preedit($objtype,$attrname,$i) "
				setAttrToLower $attrname,$i 
			"
		    }
		    multi {
			set listitems ""
			foreach li [lindex $subtype 1] {
			    if {"[lindex $li 0]" == "list"} {
				eval lappend listitems [lindex $li 1]
			    } else {
				lappend listitems $li
			    }
			}
			combobox $af.ele_$i -entries $listitems
			$af.ele_$i configure \
				-textvariable editattrs($attrname,$i)
			if {[set idx [lsearch $listitems \
					$editattrs($attrname,$i)]] >= 0} {
			    $af.ele_$i see $idx
			}
			set preedit($objtype,$attrname,$i) "
				setAttrToLower $attrname,$i 
			"
		    }
		    file {
			frame $af.ele_$i
			entry $af.ele_$i.e -font $ifn -width 40 -bg white \
				-textvariable editattrs($attrname,$i)
			button $af.ele_$i.b -font $ifn -text browse... \
			  -command [list setAttrFile f editattrs($attrname,$i)]
			pack $af.ele_$i.e -side left
			pack $af.ele_$i.b -side left -padx 4
		    }
		    dirpath {
			frame $af.ele_$i
			entry $af.ele_$i.e -font $ifn -width 40 -bg white \
				-textvariable editattrs($attrname,$i)
			button $af.ele_$i.b -font $ifn -text browse... \
			  -command [list setAttrFile p editattrs($attrname,$i)]
			pack $af.ele_$i.e -side left
			pack $af.ele_$i.b -side left -padx 4
		    }
		    textints {
			text  $af.ele_$i -font $ifn -bg white -width 30 \
				-height 5
			set postedit($objtype,$attrname,$i) "
				setAttrFromTextInts $af.ele_$i $attrname
			"
			set preedit($objtype,$attrname,$i) "
				setTextIntsFromAttr $af.ele_$i $attrname
			"
		    }
		    text {
			text  $af.ele_$i -font $ifn -bg white -width 30 \
				-height 5
			set postedit($objtype,$attrname,$i) "
				setAttrFromText $af.ele_$i $attrname
			"
			set preedit($objtype,$attrname,$i) "
				setTextFromAttr $af.ele_$i $attrname
			"
		    }
		    string {
			entry $af.ele_$i -font $ifn -width 30 -bg white \
				-textvariable editattrs($attrname,$i)
		    }
		    exprstring {
			entry $af.ele_$i -font $ifn -width 30 -bg white \
				-textvariable editattrs($attrname,$i)
		    }
		    intstring {
			entry $af.ele_$i -font $ifn -width 30 -bg white \
				-textvariable editattrs($attrname,$i)
		    }
		    double {
			entry $af.ele_$i -font $ifn -width 12 -bg white \
				-textvariable editattrs($attrname,$i)
		    }
		    int    {
			entry $af.ele_$i -font $ifn -width 10 -bg white \
				-textvariable editattrs($attrname,$i)
		    }
		    default {
			#entry $af.ele_$i -font $ifn -width 10 -bg white \
				-textvariable editattrs($attrname,$i)
		    }
		}
		pack $af.ele_$i -side left -padx 2 
	 	incr i
	    }
	    # check if a color editor should be added
	    # if attrname contains the word 'color' and exactly 3 previous
	    # sub-names were added (r g b)
	    if {[regexp -nocase color $attrname] && $i == 3} {
		pack [button $af.ele_$i -command "chooseColor $attrname" \
			-font $ifn -text "choose..."] -side left -padx 2
	    }
	}
	obstcl::delete $obj
    }
}

proc chooseColor {attr_name} {
    global editattrs 
    set r $editattrs($attr_name,0)
    set g $editattrs($attr_name,1)
    set b $editattrs($attr_name,2)
    if {[catch {set init [format #%02x%02x%02x $r $g $b]}]} {
	set init #808080
    } elseif {[string length $init] != 7} {
	set init #808080
    }
    set newcolor [tk_chooseColor -initialcolor $init -title "Choose Color"]
    if {[string length $newcolor]} {
        set r [string range $newcolor 1 2]
        set g [string range $newcolor 3 4]
        set b [string range $newcolor 5 6]
        set editattrs($attr_name,0) [format %d 0x$r]
        set editattrs($attr_name,1) [format %d 0x$g]
        set editattrs($attr_name,2) [format %d 0x$b]
    }
}

proc setTextFromAttr {textwid attr_name} {
    global editattrs editobj
    $textwid delete 1.0 end
    set lines [lrange [split [$editobj dump] \n] 1 end]
    while {[lindex $lines end] == "end" || \
	   [string length [lindex $lines end]] == 0 && [llength $lines]} {
	set lines [lrange $lines 0 end-1]
    }
    foreach line $lines {
	set line [string trim $line]
	set line [string trim $line \"]
	$textwid insert end "$line\n"
    }
}

proc setAttrFromText {textwid attr_name} {
    global editattrs
    set attrlines ""
    set spc ""
    foreach line [split [$textwid get 1.0 end] \n] {
	if {[string length $line]} {
	    append attrlines $spc "\"[string trim $line]\""
	    set spc " "
	}
    }
    set editattrs($attr_name,0) $attrlines
}

proc setTextIntsFromAttr {textwid attr_name} {
    global editattrs editobj
    $textwid delete 1.0 end
    set lines [lrange [split [$editobj dump] \n] 1 end]
    while {[lindex $lines end] == "end" || \
	   [string length [lindex $lines end]] == 0 && [llength $lines]} {
	set lines [lrange $lines 0 end-1]
    }
    foreach line $lines {
	set line [string trim $line]
	set line [string trim $line \"]
	$textwid insert end "$line\n"
    }
}

proc setAttrFromTextInts {textwid attr_name} {
    global editattrs
    set attrlines ""
    set spc ""
    foreach line [split [$textwid get 1.0 end] \n] {
	if {[string length $line]} {
	    append attrlines $spc "[string trim $line]"
	    set spc " "
	}
    }
    set editattrs($attr_name,0) $attrlines
}

proc setAttrToLower {attridx} {
    global editattrs
    set editattrs($attridx) [string tolower $editattrs($attridx)]
}

proc setAttrFile {type varname} {
    upvar #0 $varname myvar
    set fn ""
    set ftypes {{All *} {Symbol .sym} {Font .fnt} {Dbase .dbx} \
		{Shape .shp}  {Gif .gif} {Jpg .jpg} {Tiff .tif*} {Png *.png} }
    switch $type {
	f {set fn [tk_getOpenFile -filetypes $ftypes]}
	p {set fn [tk_chooseDirectory]}
    }
    # make into relative pathname if file is contained in current directory tree
    if {[string length $fn]} {
	regsub ^[pwd]/ $fn {} fn
	set myvar $fn
    }
}

proc loadHelp {} {
    global title
    tk_messageBox -title $title -message \
     "MapServer Editor\nTom Poindexter\ntpoindex@nyx.net\nCopyright 2000\n\n\
      Quickie hints and help:\n\n\
       Map object tree:\n\
       -  Primary mouse button: select map object to edit.\n\
       -  Secondary mouse button: popup object menu on selected object.\n\
       -  You can only paste a cut or copied object into valid parent\
          objects.\n\
       -  Attribute nodes are not selectable, select parent object to\
                    edit.\n\n\
       Editor notes:\n\
       - Changes to object attributes are validated and stored when switching\
         to another object,\n     or may be validated by pressing the\
         'validate/store' button.\n\
       - For map projection attributes, do not include quotes ( \" \" ).\n\
       - You may use keyboard arrow keys to scroll the projection\
                   and points attributes text box.\n\
       - To cause an attribute to be ignored, erase all attribute data, or\
         select 'default'.\n\n\
       General:\n\
       - Comments in an existing mapfile are not preserved when saved.\n\
       - Switching object attribute display on/off causes complete tree\
                    redraw.\n\
       - Loading a mapfile is much faster with object attribute display off.\n\
       - Object tree and edit panes are adjustable, drag the middle frame\
          to resize.\n\
    "
}

proc setObjFromEdit {setobj} {
    global editattrs 
    global postedit
    set objtype [$setobj getType]
    set errors ""
    foreach {attr_name oldval} [$setobj getAttr] {
	set attr_type [$setobj getAttrType $attr_name]
        set newval ""
	if {"$attr_type" == "text" || "$attr_type" == "textints"} {
	    set attridx $objtype,$attr_name,0
	    if {[string length $postedit($attridx)]} {
		catch { uplevel #0 $postedit($attridx) }
	    }
	    set newval $editattrs($attr_name,0)
	} else {
	    foreach attridx \
			[lsort [array names postedit $objtype,$attr_name,*]] {
	        if {[string length $postedit($attridx)]} {
	            catch { uplevel #0 $postedit($attridx) }
	        }
	        set attridx2 [join [lrange [split $attridx ,] 1 end] ,]
		if {[string length $editattrs($attridx2)]} {
	            lappend newval $editattrs($attridx2)
		}
	    }
	}
	$setobj setAttr $attr_name $newval
        foreach {rc msg} [$setobj validateAttr $attr_name] {break}
	if {$rc == 0} {
	    append errors "$attr_name: $msg \n\n"
	} elseif {$rc == 2} {
	    # attribute newval is null, set special flag
	    $setobj setAttr $attr_name #none
	} else {
	    # attribute parsed ok
	}
    }
    return $errors
}

proc saveEditAttrs {} {
    global edit tree editobj edittreeobj editattrs dirty title
    set oldobjvals [$editobj getAttr]
    set errors ""
    # first, instantiate a new object, check attributes validity
    set testobj [mstclobj::[$editobj getType]TclObj #auto {}]
    set errors [setObjFromEdit $testobj]
    set newobjvals [$testobj getAttr]
    obstcl::delete $testobj
    if {[string length $errors]} {
        tk_messageBox -title $title \
                  -icon error -type ok -message "Errors:\n\n$errors"
	setMsg "object [$editobj getType] has errors, not stored"
    } elseif {[string compare $newobjvals $oldobjvals] != 0} {
	set dirty 1
        setObjFromEdit $editobj
	setMsg "object [$editobj getType] attributes saved"
	$tree nodeconfigure $edittreeobj \
			-text "[$editobj getType] [$editobj getTag]"
    } else {
    }
}

proc revertEdit {} {
    loadEditAttrs
    setMsg "changes discarded"
}

proc loadEditAttrs {} {
    global edit tree editobj edittreeobj editattrs 
    global preedit
    foreach idx [array names editattrs] {
	set editattrs($idx) ""
    }
    set objtype [$editobj getType]
    foreach {attr_name oldval} [$editobj getAttr] {
	if {"$oldval" == "#none"} {
	    set oldval ""
	    foreach v [array names preedit $objtype,$attr_name,*] {
	 	lappend oldval ""
	    }	
	}
	set i 0
	foreach {val} $oldval {
	    set editattrs($attr_name,$i) $val
 	    incr i
	}
	# apply preedit tcl code
	set i 0
	foreach {val} $oldval {
	    if {[info exists preedit($objtype,$attr_name,$i)]} {
	        if {[string length $preedit($objtype,$attr_name,$i)]} {
	            catch { uplevel #0 "$preedit($objtype,$attr_name,$i)" } res
	        }
	    }
 	    incr i
	}
    }
}

proc clearEditor {} {
    global edit tree editobj edittreeobj
    if {[string length $editobj]} {
	saveEditAttrs
    }
    pack forget [pack slaves $edit]
}

proc loadEditor {treeobj} {
    global edit tree editobj edittreeobj
    set obj [$tree nodecget $treeobj -user]
    clearEditor
    set type [$obj getType]
    pack $edit.$type -fill x -expand 1 -anchor nw
    set editobj $obj
    set edittreeobj $treeobj
    loadEditAttrs
}

proc checkSelect {} {
    # only allow map objects to be selected
    global tree
    set treeobj [$tree selection get]
    set obj [$tree nodecget $treeobj -user]
    if {[string length $obj] == 0} {
	$tree selection clear
    } else {
        loadEditor $treeobj
    }
}

proc setMsg {txt} {
    .msg configure -text $txt
    update idletasks
}

proc doExit {} {
    global mapobj dirty title
    if {[string length $mapobj] && $dirty} {
	set opt [tk_messageBox -title $title \
                -icon warning -type okcancel -message "Changes\
                have been made to mapfile.\n\nExit without saving\
                current mapfile?"]
	switch $opt {
	    ok {}
	    cancel {return}
	}
    }
    exit
}

proc newMap {} {
    global mapobj mapfile tree dirty editobj edittreeobj title treeObjId
    if {[string length $mapobj] && $dirty} {
	set opt [tk_messageBox -icon warning -title $title \
               -type okcancel -message "Changes\
                have been made to mapfile.\n\nLoad new mapfile without saving\
                current mapfile?"]
	switch $opt {
	    ok {}
	    cancel {return}
	}
    }
    catch {obstcl::delete $mapobj}
    catch {$tree delete 0; $tree configure -text "(no mapfile opened)"}
    set mapobj ""
    set mapobj [mstclobj::mapTclObj #auto]
    set treeObjId 0
    $tree configure -text "new mapfile"
    $tree insert 0 -text "map" -image img_map -user $mapobj
    populateTree $tree $mapobj 0
    $tree see      0
    $tree expand   0
    nodebind $tree 0
    setMsg "new mapfile object"
    set mapfile ""
    set dirty 0
    set editobj ""
    set edittreeobj ""
    loadBanner
    set opt [tk_messageBox -icon info -title $title  -type okcancel -message \
	"You should save you new mapfile immediately, in order to ensure\
         \nyour data and symbol directory paths are relative paths.\n\n\
	 \nSave your new mapfile now?" ]
    if {"$opt" == "ok"} {
	saveMap
    }
}

proc nodebind {tree treeobj} {
    $tree nodebind $treeobj <ButtonRelease-2> \
			[list +doPopup $treeobj %X %Y]
    $tree nodebind $treeobj <ButtonRelease-3> \
			[list +doPopup $treeobj %X %Y]
}

proc saveMap {} {
    global mapobj mapfile tree dirty editobj title
    if {[string length $mapobj] == 0} {
	tk_messageBox -icon info -type ok -title $title \
            -message "No map objects to save.\
            \n\nYou should begin a new mapfile, or open an existing one first."
	return
    }
    if {[string length $editobj]} {
	saveEditAttrs
    }
    if {[string length $mapobj] > 0 && [string length $mapfile] == 0} {
	saveMapAs
	return
    }
    # backup previous map, if any
    if {[file exists $mapfile]} {
	if {[catch {file copy -force -- $mapfile $mapfile.bak}]} {
	    set bakmsg ", warning: previous file could not be saved\
                  to [file tail $mapfile].bak"
	} else {
	    set bakmsg ", previous file saved as [file tail $mapfile].bak"
	}
    } else {
	set bakmsg ""
    }
    $mapobj dumpToFile $mapfile
    setMsg "mapfile saved to $mapfile$bakmsg"
    $tree configure -text [file tail $mapfile]
    set dirty 0
}

proc saveMapAs {} {
    global mapobj mapfile maptypes title
    if {[string length $mapobj] == 0} {
	tk_messageBox -icon info -type ok -title $title \
            -message "No map objects to save.\
            \n\nYou should begin a new mapfile, or open an existing one first."
	return
    }
    set fn [tk_getSaveFile -defaultextension .map -filetypes $maptypes]
    if {"$fn" != ""} {
	set mapfile $fn
        catch {cd [file dirname $mapfile]}
	saveMap 
    } else {
	setMsg "map not saved"
    }
}

proc openMap {} {
    global maptypes mapobj mapfile tree dirty title treeObjId
    if {[string length $mapobj] && $dirty} {
	set opt [tk_messageBox -icon warning -title $title \
              -type okcancel -message "Changes\
                have been made to mapfile.\n\nLoad new mapfile without saving\
                current mapfile?"]
	switch $opt {
	    ok {}
	    cancel {return}
	}
    }
    catch {obstcl::delete $mapobj}
    set treeObjId 0
    catch {$tree delete 0; $tree configure -text "(no mapfile opened)"}
    set mapobj ""

    set mapfile [tk_getOpenFile -filetypes $maptypes -parent .]
    if {$mapfile == {}} {
	return
    }
    catch {cd [file dirname $mapfile]}
    openMapFile
}

proc openMapFile {} {      
    global maptypes mapobj mapfile tree dirty editobj edtittreeobj treeObjId
    set lexer    [mstclobj::mapLexer #auto]
    if {[catch {$lexer initFromFile $mapfile} result]} {
	setMsg $result
	return
    }
    set mapobj [mstclobj::mapTclObj #auto]
    if {[catch {$mapobj parse $lexer} result]} {
	obstcl::delete $mapobj
	obstcl::delete $lexer
	setMsg $result
	return
    }
    $tree configure -text [file tail $mapfile]
    set mapname [$mapobj getTag]
    if {"$mapname" == "#none"} {set mapname ""}
    set treeObjId 0
    $tree insert 0 -text "map $mapname" -image img_map -user $mapobj
    setMsg "loading map object..."
    populateTree $tree $mapobj 0
    $tree see      0
    $tree expand   0
    nodebind $tree 0
    obstcl::delete $lexer
    setMsg "[file tail $mapfile] loaded"
    set dirty 0
    set editobj ""
    set edittreeobj ""
    loadBanner
}

proc redrawTree {} {
    global mapobj mapfile tree treeObjId 
    if {[string length $mapobj]} {
	setMsg "standby..."
	set treetext [$tree cget -text]
        set treeObjId 0
        catch {$tree delete $treeObjId}
        $tree configure -text $treetext
        $tree insert 0 -text "map" -image img_map -user $mapobj
        populateTree $tree $mapobj 0
        $tree see      0
        $tree expand   0
        nodebind $tree 0
	setMsg "ready"
    }
}

proc populateTree {tree obj parent} {
    global showattrnodes treeObjId node_img
    if {$showattrnodes} {
        foreach {attr val} [$obj getAttr] {
            $tree insert [incr treeObjId] -text $attr -parent $parent \
		-image img__attr -user $obj
        }
    }
    foreach {sub} [$obj getSubObjs] {
	set type [$sub getType]
        if {[info exists node_img($type)]} {
	    set img img_$type
	} else {
	    set img img__default
	}
	$tree insert [incr treeObjId] -text "[$sub getType] [$sub getTag]" \
			     -parent $parent -image $img -user $sub
        nodebind $tree $treeObjId
	setMsg "loading [$sub getType] object..."
	populateTree $tree $sub $treeObjId
    }
}

proc mkPopup {treeobj} {
    global tree cutbuffer
    set obj [$tree nodecget $treeobj -user]
    catch {rename .pop.add {}}
    catch {rename .pop     {}}
    menu .pop -tearoff 0
    .pop add command -label "expand all"   -command "$tree expand   $treeobj 1"
    .pop add command -label "collapse all" -command "$tree collapse $treeobj 1"
    .pop add separator
    .pop add command -label copy     -command "copyObj  $treeobj"
    .pop add command -label cut      -command "cutObj   $treeobj"
    .pop add command -label paste    -command "pasteObj $treeobj"
    if {"[$obj getType]" == "map"} {
	# no cut/copy functions for top level map obj,use file methods instead
	.pop entryconfigure 3 -state disabled
	.pop entryconfigure 4 -state disabled
    }

    # check if cutbuffer empty, diable paste if is empty
    if {![string length $cutbuffer]} {
	.pop entryconfigure 5 -state disabled
	.pop entryconfigure 5 -label paste
    } else {
        set cut_type [lindex $cutbuffer 0]
        set cut_text [lindex $cutbuffer 1]
	.pop entryconfigure 5 -label "paste [lindex $cutbuffer 0] object" \
		-state normal
        if {[lsearch [$obj getValidSubObjs] $cut_type] == -1} {
	     .pop entryconfigure 5 -state disabled
	}
    }

    set sublist [$obj getValidSubObjs]
    if {[string length $sublist]} {
        menu .pop.add -tearoff 0
	foreach objtype $sublist {
            .pop.add add command -label $objtype \
			-command "insertObj $treeobj $objtype "
	}

        .pop add separator
        .pop add cascade -label "insert object" -menu .pop.add
    }
}

proc doPopup {treeobj x y} {
    global tree mapobj
    if {"[$tree selection get]" != "$treeobj"} {
	return
    }
    mkPopup $treeobj
    tk_popup .pop $x $y
}

proc copyObj {treeobj} {
    global cutbuffer
    set obj [$tree nodecget $treeobj -user]
    setMsg "object [$obj getType] [$obj getTag] copied"
    set cutbuffer [list [$obj getType] [$obj dump]]
}

proc cutObj {treeobj} {
    global tree cutbuffer
    global editobj
    set obj [$tree nodecget $treeobj -user]
    set cutbuffer [list [$obj getType] [$obj dump]]
    setMsg "object [$obj getType] [$obj getTag] deleted"
    $tree delete $treeobj   
    obstcl::delete $obj
    set editobj ""
    clearEditor
}

proc pasteObj {treeobj} {
    global cutbuffer tree dirty treeObjId
    set obj [$tree nodecget $treeobj -user]
    if {![string length $cutbuffer]} {
	setMsg "object cutbuffer empty"
        return
    }    
    set cut_type [lindex $cutbuffer 0]
    set cut_text [lindex $cutbuffer 1]
    if {[lsearch [$obj getValidSubObjs] $cut_type] == -1} {
	setMsg "object cutbuffer type $cut_type, valid [$obj getType]\
               sub-objects: [$obj getValidSubObjs]"
	return
    }
    set newobj [mstclobj::${cut_type}TclObj #auto $obj]
    set lexer [mstclobj::mapLexer #auto]
    $lexer initFromStr $cut_text
    $newobj parse $lexer
    obstcl::delete $lexer
    set type [$newobj getType]
    if {[info exists node_img($type)]} {
	set img img_$type
    } else {
	set img img__default
    }
    $tree insert [incr treeObjId] -text "[$newobj getType] [$newobj getTag]" \
		-parent $treeobj -image $img
    set newTreeObj $treeObjId
    nodebind $tree $newTreeObj
    populateTree $tree $newobj $newTreeObj
    $tree selection clear
    $tree see       $newTreeObj
    $tree expand    $newTreeObj
    setMsg "pasted object $cut_type into [$obj getType] [$obj getTag]"
    set dirty 1
}

proc insertObj {treeobj newobjtype} {
    global tree dirty treeObjId node_img
    set obj [$tree nodecget $treeobj -user]
    set newobj [mstclobj::${newobjtype}TclObj #auto $obj]
    if {[info exists node_img($newobjtype)]} {
	set img img_$newobjtype
    } else {
	set img img__default
    }
    $tree insert [incr treeObjId] -text [$newobj getType]  -parent $treeobj \
		-image $img -user $newobj
    set newTreeObj $treeObjId
    nodebind $tree $newTreeObj
    populateTree $tree $newobj $treeObjId
    $tree selection clear
    $tree see       $newTreeObj
    $tree expand    $newTreeObj
    setMsg "inserted new [$newobj getType]"
    set dirty 1
}

proc parseWithMS {} {
    global mapobj mapfile dirty title

    if {[catch {package require Mapscript}]} {
	tk_messageBox -icon error -type ok -title $title \
            -message "Mapscript Tcl extension is not available.  "
	return
    }

    if {[string length $mapobj] == 0} {
	tk_messageBox -icon warning -type ok -title $title \
            -message "No mapfile is\
             loaded.  Open an existing mapfile in order to use Mapscript parser"
	return
    }
    if {[string length $mapfile] == 0} {
	tk_messageBox -icon warning -type ok -title $title \
            -message "You must save your\
             current map objects to a file before you can use MapServer parser"
	return
    }
    if {[string length $mapobj] && $dirty} {
	set opt [tk_messageBox -icon warning -title $title \
             -type okcancel -message "Changes\
                have been made to mapfile.\n\nParse existing mapfile without\
                saving current mapfile objects?"]
	switch $opt {
	    ok {}
	    cancel {return}
	}
    }

    set mapParseObj [mapscript::mapObj -args $mapfile]
    if {"$mapParseObj" == "NULL"} {
        set msg "Mapfile [file tail $mapfile] has errors:\n\n"
        foreach v {-code -routine -message} {
           append msg "$v: [mapscript::ms_error cget $v]\n"
        }
    } else {
	set msg "Mapfile [file tail $mapfile] parsed successfully"
    }
    # destroy map object
    catch {rename $mapParseObj {}}

    tk_messageBox -icon info -type ok -title $title \
	-message $msg
    
}

proc getInterps {} {
    global tcl_platform

    if {"$tcl_platform(platform)" == "windows"} {
	update
        set all_interps [dde services TclEval {}]
        set interps ""
        foreach intrp $all_interps {
            lappend interps [lindex $intrp 1]
        }
    } else {
        set interps [winfo interps]
    }
    return $interps
}


proc startMSExe {exe_name} {
    global mapobj mapfile startupdir tcl_platform title
    if {[lsearch [getInterps] $exe_name] == -1} {
        if {"$tcl_platform(platform)" == "windows"} {
            set start_exe [list exec start [file join $startupdir $exe_name] &]
            set error_msg "Cannot 'send' to $exe_name via DDE emulation."
        } else {
            set start_exe [list exec [file join $startupdir $exe_name] &]
            set error_msg "Cannot 'send' to $exe_name. Check your xhost\
                settings, xhost should report\
                \n\"access control enabled, only authorized clients can\
                 connect\" "
        }
        if {[catch $start_exe result]} {
            tk_messageBox -icon error -type ok \
                -title $title \
                -message "Cannot start $exe_name\n\n$result"
            return 0
        } else {
            setMsg "please wait, starting $exe_name"
            # wait for Tk program to startup
            set i 20
            while {$i >= 0 && [lsearch [getInterps] $exe_name] == -1} {
                update
                after 1000
                incr i -1
                update
            }
            if {$i <= 0 && [lsearch [getInterps] $exe_name] == -1} {
                tk_messageBox -icon error -type ok \
                     -title $title \
                     -message "$error_msg\n\n$result"
                setMsg "could not contact $exe_name"
                return 0
            }
            after 1000
        }
    }
    update
    if {[catch {send $exe_name update} result]} {
        tk_messageBox -icon error -type ok -title $title \
                     -message "$error_msg\n\n$result"
        return 0
    }

    return 1
}

proc startMShapex {} {
    global mapobj mapfile dirty startupdir title
    set exe_name msshapex.tcl
    if {! [startMSExe $exe_name]} {
        return
    }
    setMsg "ready"
}

proc startMSex {} {
    global mapobj mapfile dirty startupdir title
    set exe_name msexplorer.tcl

    if {[string length $mapfile] == 0} {
	tk_messageBox -icon warning -type ok -title $title \
          -message "You must save your\
           current map objects to a file before you can use MapServer Explorer"
	return
    }
    if {[string length $mapobj] && $dirty} {
	set opt [tk_messageBox -icon warning -title $title \
            -type okcancel -message "Changes\
                have been made to mapfile.\n\nView existing mapfile without\
                saving current mapfile?"]
	switch $opt {
	    ok {}
	    cancel {return}
	}
    }
    if {! [startMSExe $exe_name]} {
        return
    }

    update
    if {[catch {send $exe_name remoteOpenFile $mapfile} result]} {
	tk_messageBox -icon error -type ok -title $title \
		     -message "Cannot 'send' to $exe_name:\n\n$result"
    }
    setMsg "ready"
}

proc remoteOpenFile {filename} {
    global maptypes mapobj mapfile tree dirty title
    wm deiconify .
    raise .

    if {[file isfile $filename]} {
        if {[string length $mapobj] && $dirty} {
	    set opt [tk_messageBox -icon warning -title $title \
                -type okcancel \
		-message "Changes\
                 have been made to mapfile.\n\nLoad new mapfile without saving\
                 current mapfile?"]
	    switch $opt {
	        ok {}
	        cancel {return}
	    }
        }
    } else {
	tk_messageBox -icon error -type ok -title $title \
                -message "File '$filename' is not found."
        return

    }
    catch {obstcl::delete $mapobj}
    catch {$tree delete $mapobj; $tree configure -text "(no mapfile opened)"}
    set mapobj ""

    set mapfile $filename
    catch {cd [file dirname $mapfile]}
    openMapFile
}




############################################################################
# 'send' emulation for windows
#
# written by George A. Howlett, 
# scarfed from a comp.lang.tcl posting by George.
#
#

# --------------------------------------------------------------------------
#
# SendInit --
#	
#	Creates a "send" proc to replace the former Tk send command.
#	Uses DDE services to simulate the transfer.  This must be
#	called before any drag&drop targets are registered. Otherwise
#	they will pick up the wrong application name.
#
#	The first trick is to determine a unique application name. This
#	is what other applications will use to send to us.  Tk used to
#	do this for us.  
#	
#	Note that we can generate the same name for two different Tk 
#	applications.  This can happen if two Tk applications picking
#	names at exactly the same time.   [In the future, we should 
#	probably generate a name based upon a global system value, such
#	as the handle of the main window "."]   The proc "SendVerify"
#	below will verify that you have only one DDE server registered
#	with this application's name.
#	
# Arguments:
#	myInterp	Sets the application name explicitly to this 
#			string. If the argument isn't given, or is the
#			empty string, then the routine picks a name for
#			us.
#
# Results:
#	Returns the name of the application.
#
# Side Effects:
#	Sets the name of our application.  You can call "tk appname" to
#	get the name.  A DDE topic using the same name is also created.
#	A send proc is also automatically created.  Be careful that you
#	don't overwrite an existing send command.
#
# --------------------------------------------------------------------------

proc SendInit { {myInterp ""} } {

    # Load the DDE package.
    package require dde

    if { $myInterp == "" } {

	# Pick a unique application name, replicating what Tk used to do.
	# This is what other applications will use to "send" to us. We'll
	# use DDE topics to represent interpreters.

	set appName [tk appname]
	set count 0
	set suffix {}
	
	# Keep generating interpreter names by suffix-ing the original
	# application name with " #number".  Sooner of later we'll find
	# one that's not currently use. 

	while { 1 } {
	    set myInterp "${appName}${suffix}"
	    set myServer [list TclEval $myInterp]
	    if { [lsearch [dde services TclEval {}] $myServer] < 0 } {
		break
	    }
	    incr count
	    set suffix " \#$count"
	}
    }
    tk appname $myInterp
    dde servername $myInterp
    proc send { interp args } {
	dde eval $interp $args
    }
    return $myInterp
}


# --------------------------------------------------------------------------
#
# SendVerify --
#	
#	Verifies that application name picked is uniquely registered
#	as a DDE server.  This checks that two Tk applications don't
#	accidently use the same name.
#
# Arguments:
#	None		Used the current application name.
#
# Results:
#	Generates an error if either a server can't be found or more
#	than one server is registered.
#
# --------------------------------------------------------------------------

proc SendVerify {} {
    # Load the DDE package.
    package require dde

    set count 0
    set appName [tk appname]
    foreach server [dde services TclEval {}] {
	set topic [lindex $server 1]
	if { [string compare $topic $appName] == 0 } {
	    incr count
	}
    }
    if {$count == 0} {
	error "Service not found: wrong name registered???"
    } 
    if { $count > 1 } {
	error "Duplicate names found for \"[tk appname]\""
    } 
}

#
###########################################################################


# main code


if {"$tcl_platform(platform)" == "windows"} {
    set scriptname [file tail $scriptname]
    SendInit $scriptname
    if {[catch {SendVerify} reason]} {
        tk_messageBox -icon error -type warning \
                -title $title \
                -message "Cannot initialize \"send\" emulation using DDE.\n\
                    \n$reason\n\
                    \nInteraction with other MapServer Workbench tools\
                    may not function properly."
    }
}


mkMain
setMsg "standby, building object editors..."
update idletasks
mkEditors

setMsg "ready"

