#!/bin/sh
# The next line is executed by /bin/sh, but not Tcl \
    exec /usr/local/bin/wish8.3 $0 ${1+"$@"}

# if you are installing by hand
# replace /usr/local/bin/wish8.3
# on line 3 with the full path to your wish executable, verison 8.3 or higher.

global startupdir
set startupdir ""
set scriptname [info script]
# on unix, chase symbolic links until we get to the real program
if {"$tcl_platform(platform)" == "unix"} {
    while {[catch {set scriptname [file readlink $scriptname]}] == 0} { }
}
set startupdir [file dirname $scriptname]
if {"$startupdir" == "."} {
     set startupdir [pwd]
}
lappend auto_path $startupdir [file dirname $startupdir]


package require Mapscript

if {[info exists env(TEMP)]} {
  set tmpdir $env(TEMP)
} else {
  # Urgh
  set tmpdir /tmp
  catch {file mkdir /tmp}
}

option add *Dialog.msg.wrapLength 700
option add *Dialog.msg.font {helvetica 12 bold}

global title
set title "MapServer Explorer"

global viewfunction
set viewfunction zoomin

global zoomvalue
set zoomvalue 2.0

global mapTypes
set mapTypes {
 {{Map Files}       {.map}        }
 {{All Files}        *            }
}

global mapFile
set mapFile ""

global banner_uu
set banner_uu \
{R0lGODdhWAKQAYAAAPj8+AAAACwAAAAAWAKQAQAC/oSPqcvtD6OctNqLs968+w+G4kiW5omm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}


proc die {msg} {
    global mapFile title
    set msg "Mapfile [file tail $mapFile] has errors:\n\n$msg\n\n"
    foreach v {-code -routine -message} {
	append msg "$v: [mapscript::ms_error cget $v]\n"
    }
    tk_messageBox -icon info -title $title \
        -type ok -message $msg
}

# Lay out the UI
proc setupWindow {{w .}} {
   global title
   global config
   global data

   global banner_uu
   
   set config(state)       "initUI"
   set config(mode)        zoomin
   set config(scalefactor) 2

   if {$w == "."} {
      set f ""
   } else {
      set f $w
   }
   wm title . $title
   wm protocol . WM_DELETE_WINDOW doExit

   menu .menuBar -tearoff 0

   .menuBar add cascade -menu .menuBar.file -label "File" -underline 0
   .menuBar add cascade -menu .menuBar.view -label "View" -underline 0
   .menuBar add cascade -menu .menuBar.tool -label "Tools" -underline 0
   .menuBar add cascade -menu .menuBar.help -label "Help"  -underline 0


   menu .menuBar.file -tearoff 0
   .menuBar.file add command -label "Open"     -underline 0 \
		-command "openMap ." -accelerator Ctrl+O
   bind all <Control-KeyPress-o> +openMap

   .menuBar.file add separator
   .menuBar.file add command -label "Quit"     -underline 0 -command doExit \
		-accelerator Ctrl+Q
   bind all <Control-KeyPress-q> +doExit 

   menu .menuBar.view -tearoff 1
   menu .menuBar.view.zoom -tearoff 0
   .menuBar.view.zoom add radiobutton -label " 1.5x " \
		-command setZoom -variable zoomvalue -value 1.5
   .menuBar.view.zoom add radiobutton -label " 2x " \
		-command setZoom -variable zoomvalue -value 2.0
   .menuBar.view.zoom add radiobutton -label " 4x " \
		-command setZoom -variable zoomvalue -value 4.0
   .menuBar.view.zoom add radiobutton -label " 8x " \
		-command setZoom -variable zoomvalue -value 8.0
   .menuBar.view.zoom add radiobutton -label " 16x " \
		-command setZoom -variable zoomvalue -value 16.0
   .menuBar.view.zoom add radiobutton -label " 32x " \
		-command setZoom -variable zoomvalue -value 32.0

   .menuBar.view add radiobutton -label "Zoom in" \
                -command setFunction -variable viewfunction -value zoomin
   .menuBar.view add radiobutton -label "Zoom out" \
		-accelerator Shift+Button1 \
                -command setFunction -variable viewfunction -value zoomout
   .menuBar.view add cascade -label "Zoom value..." -menu .menuBar.view.zoom 
   .menuBar.view add separator
   .menuBar.view add radiobutton -label "Pan" -accelerator Ctrl+Button1 \
                -command setFunction -variable viewfunction -value pan
   .menuBar.view add separator
   .menuBar.view add radiobutton -label "Query" -accelerator Button2 \
                -command setFunction -variable viewfunction -value query
   .menuBar.view add separator
   .menuBar.view add command -label "Reset to map extents" \
                -command "extents ." -accelerator Ctrl+R
   bind all <Control-KeyPress-r> +extents

   menu .menuBar.tool -tearoff 0
   .menuBar.tool add command -label "Start MapServer Editor" \
                -command startMSed -underline 0 -accelerator Ctrl+E
   bind all <Control-KeyPress-e> startMSed

    menu .menuBar.help -tearoff 0
    .menuBar.help add command -label "About" -command loadHelp -underline 0


   .   configure -menu .menuBar

   image create photo $f.mapimage -data $banner_uu
   set config(image) $f.mapimage
   canvas $f.img -bd 0 -width 600 -height 400
   $f.img create image 0 0 -image $f.mapimage -anchor nw
   set config(canvas) $f.img

   image create photo $f.scaleimage
   set config(scale) $f.scaleimage
   label $f.scale     -bd 1 -relief flat -image $f.scaleimage

   frame $f.rhs
   image create photo $f.legendimage
   set config(legend) $f.legendimage
   label $f.rhs.leg_lab   -text Legend -anchor n
   label $f.rhs.legend    -bd 1 -relief flat -image $f.legendimage 
   label $f.rhs.location  -bd 1 -relief flat -textvariable config(location)

   label $f.rhs.lay_lab   -text Layers
   listbox $f.rhs.layers -width 25 -height 8 -selectmode multiple \
		-exportselection 0 -yscrollcommand "$f.rhs.layerscroll set"
   scrollbar $f.rhs.layerscroll  -command "$f.rhs.layers  yview"
   set config(layers) $f.rhs.layers
   button      $f.rhs.refresh -text "Refresh" -command "drawImages $w"
   
   label $f.msg -text "" -anchor w

   grid $f.img    -row 0 -column 0    -in $w -sticky nsew -padx 5 -pady 5
   grid $f.scale  -row 1 -column 0    -in $w -sticky ew -pady 5

   grid $f.rhs.leg_lab -row 0 -column 0    -in $f.rhs -sticky new
   grid $f.rhs.legend  -row 1 -column 0    -in $f.rhs -sticky new
   grid $f.rhs.lay_lab -row 2 -column 0    -in $f.rhs -sticky sew
   grid $f.rhs.layers  -row 3 -column 0    -in $f.rhs -sticky sew
   grid $f.rhs.layerscroll -row 3 -column 1    -in $f.rhs -sticky nsw
   grid $f.rhs.refresh -row 4 -column 0    -in $f.rhs
   grid $f.rhs.location -row 5 -column 0    -in $f.rhs -sticky nsw
   grid columnconfigure $f.rhs 0 -weight 1
   grid rowconfigure $f.rhs 0 -weight 1
   grid $f.rhs         -row 0 -column 1 -in $w -sticky new  -rowspan 2

   grid $f.msg        -row 3 -column 0 -in $w -sticky nsew -columnspan 2

   grid columnconfigure $w 0 -weight 1
   grid rowconfigure $w 0 -weight 1

   bind $f.img <Enter>  {}
   bind $f.img <Leave>  "set config(location) {}"
   bind $f.img <1>      "Btn1Down $w %W %x %y"
   bind $f.img <Shift-1> "BtnShift1Down $w %W %x %y"
   bind $f.img <Control-1> "BtnControl1Down $w %W %x %y"
   bind $f.img <2> "Btn23Down $w %W %x %y"
   bind $f.img <3> "Btn23Down $w %W %x %y"

   bind $f.img <Motion> "Motion   $w %W %x %y"

   setFunction
}

proc loadHelp {} {
    global title
    tk_messageBox -title $title -message "\
      MapServer Explorer\nTom Poindexter\ntpoindex@nyx.net\nCopyright 2000\n\
      Based on \"maptcl-demo.tcl\" by Peter deVoil, Copyright 1999\n\n\
      Quickie hints and help:\n\n\
      - 'View' menu may be torn-off, select the dashed line on the menu.\n\
      - You must select at least one layer to view after loading a mapfile.\n\
      - Query only considers layers currently viewed, and only if a query\
         attribute\n     is defined for that layer.\n\
    "
}

proc setFunction {{w .}} {
    global config
    global viewfunction
    if {$w == "."} {
       set w ""
    }
   
    set txt "unknown"
    set config(mode) $viewfunction
    switch -- $viewfunction {
	zoomin   {$w.img configure -cursor based_arrow_down; set txt "zoom in"}
	zoomout  {$w.img configure -cursor based_arrow_up;  set txt "zoom out"}
	pan      {$w.img configure -cursor fleur; set txt "pan"}
	query    {$w.img configure -cursor question_arrow; set txt "query"}
    }
    setMsg "$txt mode selected"
}

proc setZoom {{w .}} {
    global config
    global zoomvalue
    if {$w == "."} {
       set w ""
    }
    set config(scalefactor) $zoomvalue
}

proc doExit {} {
    global mapFile title tmpdir
    catch {file delete [file join $tmpdir demo[pid].gif]}
    catch {file delete [file join $tmpdir legend[pid].gif]}
    catch {file delete [file join $tmpdir scale[pid].gif]}
    exit
}

# probe map for default extents, layer names.
proc setupMap {{w .}} {
    global mapFile
    global config

    if {$w == "."} {
       set w ""
    }

    set config(state) "startMap"

    set map [mapscript::mapObj -args $mapFile]
    if {$map == "NULL"} {
	die new_mapObj
        return ""
    }

    set rectPtr [$map cget -extent]
    foreach {i} {minx miny maxx maxy} {
       set config($i,extent) [mapscript::rectObjRef $rectPtr cget -$i]
       set config($i) $config($i,extent)
    }

    $config(layers) delete 0 end
    set names ""
    for {set i 0} {$i < [$map cget -numlayers]} {incr i} {
	   set name [mapscript::layerObjRef  [$map getLayer $i] cget -name]
	   lappend names $name
    }
    set names [lsort -unique $names]
    set i 0
    foreach name $names {
	$config(layers) insert end $name
        $config(layers) selection set $i
        incr i
    }
    set config(state) "doneSetup"
    set config(map) $map
    return $map
}

proc destroyMap {{w .}} {
   global mapFile zoomvalue viewfunction
   global config
    if {$w == "."} {
       set w ""
    }
   rename $config(map) {}
   
   $config(layers) delete 0 end
   set config(map) ""
   set mapFile ""
   set zoomvalue 2.0
   set viewfunction zoomin
}

# draw images, update the UI
proc drawImages {{w .} args} {
    global config
    global mapFile
    global tmpdir 

    if {$w == "."} {
       set w ""
    }
    set config(state) "drawing"
    set map $config(map)

    $map config -width [winfo width $config(canvas)]
    $map config -height [winfo height $config(canvas)]

    set rectPtr [$map cget -extent]
    foreach {i} {minx miny maxx maxy} {
       mapscript::rectObjRef $rectPtr configure -$i $config($i) 
    }

    setMsg "preparing to draw images..."
    set img [$map prepareImage]
    if {$img == "NULL"} {die prepareImg; return}

    set layers {}
    foreach j [$config(layers) curselection] {
       lappend layers [$config(layers) get $j]
    }

    for {set i 0} {$i < [$map cget -numlayers]} {incr i} {
       set layer [$map getLayer $i]
       if {$layer == "NULL"} {die mapObj_getLayer; return}
	   set name [mapscript::layerObjRef $layer cget -name]
	   if {[lsearch $layers $name] >= 0} {
	      mapscript::layerObjRef $layer configure -status $mapscript::MS_ON
	   } else {
	      mapscript::layerObjRef $layer configure -status $mapscript::MS_OFF
	   }
	   setMsg "drawing $name layer..."
	   mapscript::layerObjRef $layer draw $map $img
	}

    setMsg "drawing labels..."
    $map drawLabelCache $img
    mapscript::labelCacheObjRef [$map cget -labelcache] freeCache

    mapscript::msSaveImage $img "[file join $tmpdir demo[pid].gif]" 1 1
    mapscript::msFreeImage $img
    $config(image) blank
    $config(image) config  -file [file join $tmpdir demo[pid].gif]

    setMsg "drawing legend..."
    set img [$map drawLegend]
    if {$img == "NULL"} {die mapObj_drawLegend; return}
    mapscript::msSaveImage $img [file join $tmpdir legend[pid].gif] 1 1 
    mapscript::msFreeImage $img
    $config(legend) blank
    $config(legend) config -file [file join $tmpdir legend[pid].gif]

    setMsg "drawing scalebar..."
    set img [$map drawScalebar]
    if {$img == "NULL"} {die mapObj_drawScalebar; return}
    mapscript::msSaveImage $img [file join $tmpdir scale[pid].gif] 1 1 
    mapscript::msFreeImage $img
    $config(scale) blank
    $config(scale) config -file [file join $tmpdir scale[pid].gif]

    set config(state) "ready"
    setMsg "drawing finished"
}

# A button click somewhere
proc Btn1Down {w c x y} {
  global config
  if { $config(state) == "ready" } {
    $config(mode) $w $c $x $y
  }
}


# A button click somewhere
proc BtnShift1Down {w c x y} {
  global config
  if { $config(state) == "ready" } {
    zoomout $w $c $x $y
  }
}

# A button click somewhere
proc BtnControl1Down {w c x y} {
  global config
  if { $config(state) == "ready" } {
    pan $w $c $x $y
  }
}

# A button click somewhere
proc Btn23Down {w c x y} {
  global config
  if { $config(state) == "ready" } {
    query $w $c $x $y
  }
}

# Mouse movement
proc Motion {w c x y} {
  global config
  if {$config(state) != "ready"} {
    return
  }
  set p [canvas2point $c $config(map) $x $y]
  set config(location) "[format %.2f [mapscript::pointObjRef $p cget -x]],\
	 [format %.2f [mapscript::pointObjRef $p cget -y]]"
}

# translate pixel units to map coordinate system
proc canvas2point {c map x y} {
   set mapext [mapscript::mapObjRef $map cget -extent]

   foreach {i e} {0 minx 1 miny 2 maxx 3 maxy} {
      set extent($i) [mapscript::rectObjRef $mapext cget -$e]
   }
   
   set cs [mapscript::mapObjRef $map cget -cellsize]
   
   set ix [$c canvasx $x]
   set iy [$c canvasy $y]
      
   set x [expr $extent(0) + $cs*$ix]  ;# change x,y from image to map 
				       #coordinates, offset from LL corner 
                                       #of previous image
   set y [expr $extent(3) - $cs*$iy]

   set p [mapscript::new_pointObj]
   mapscript::pointObjRef $p configure -x $x
   mapscript::pointObjRef $p configure -y $y
   return $p
}

proc zoomin {w c x y} {
   global config
   moveabout $w $c $x $y $config(scalefactor)
}
proc zoomout {w c x y} {
   global config
   moveabout $w $c $x $y [expr 1.0 / $config(scalefactor)]
}
proc pan {w c x y} {
   moveabout $w $c $x $y 1.0
}

# Adjust drawn extents of map
proc moveabout {w c x y factor} {
   global config

   set mapext [mapscript::mapObjRef $config(map) cget -extent]

   foreach {i e} {0 minx 1 miny 2 maxx 3 maxy} {
      set extent($i) [mapscript::rectObjRef $mapext cget -$e]
   }

   set cs [mapscript::mapObjRef $config(map) cget -cellsize]

   set ix [$c canvasx $x]
   set iy [$c canvasy $y]
      
   set mx [expr $extent(0) + $cs*$ix]  ;# change x,y from image to map 
                                        #coordinates, offset from LL corner 
                                        #of previous image
   set my [expr $extent(3) - $cs*$iy]

   set config(minx) [expr $mx - 0.5 * (($extent(2)-$extent(0))/$factor)]
   set config(maxx) [expr $mx + 0.5 * (($extent(2)-$extent(0))/$factor)]
   set config(miny) [expr $my - 0.5 * (($extent(3)-$extent(1))/$factor)]
   set config(maxy) [expr $my + 0.5 * (($extent(3)-$extent(1))/$factor)]

   foreach i {minx maxx miny maxy} {
      mapscript::rectObjRef $mapext configure -$i $config($i)
   }
   drawImages $w
}

proc extents {w} {
   global config

   foreach {i} {minx miny maxx maxy} {
      set config($i)  $config($i,extent)
   }
   drawImages $w 
}

proc remoteGetExtents {{w .}} {
    global config

    set extents ""
    foreach {i} {minx miny maxx maxy} {
	lappend extents $config($i)
    }
    return $extents
}

# Generate a query. 
proc query {w c x y} {
   global config

   set i 0
   while { [winfo exists .x$i] } {
     incr i
   }
   set q .x$i
   toplevel $q

   set txt [text $q.t1 -yscrollcommand "$q.s1 set"]
   scrollbar $q.s1 -command "$q.t1 yview"
   grid $q.t1 -in $q -row 0 -column 0 -sticky nsew
   grid $q.s1 -in $q -row 0 -column 1 -sticky nse
   grid columnconfigure $q 0 -weight 1
   grid rowconfigure $q 0 -weight 1
   focus $q.t1

   set p [canvas2point $c $config(map) $x $y]
   set mx [mapscript::pointObjRef $p cget -x]
   set my [mapscript::pointObjRef $p cget -y]
   $txt insert end "Query at point [format %.2f $mx],[format %.2f $my]\n"

   foreach j [$config(layers) curselection] {
       set layerName [$config(layers) get $j]
       set layer [mapscript::mapObjRef $config(map) getLayerByName $layerName]

       $txt insert end "layer '$layerName':\n"
   

       if {[set qf [mapscript::layerObjRef $layer queryUsingPoint \
					$config(map) $p 1 1]]  != "NULL"} {
          foreach {layName name_values} \
				[mapscript::getDBFQueryInfo $config(map) $qf] {
	     $txt insert end "found in layer '$layName':\n"
             foreach {item value} $name_values {
                $txt insert end " '$item' - '$value'\n"
             }
          }
       }
    }
}

proc openMap {w} {
   global mapTypes mapFile title
   set mapFile [tk_getOpenFile -filetypes $mapTypes -parent $w]
   if {$mapFile == {}} { return }

   if {"[setupMap $w]" != ""} {
       update
       drawImages $w
       setMsg "mapfile [file tail $mapFile] loaded, select layers to view\
           and Refresh"
    } else {
        setMsg "mapfile [file tail $mapFile] could not be loaded"
   }
}

proc remoteOpenFile {filename} {
    global mapFile title

    wm deiconify .
    raise .
    if {[file isfile $filename]} {
	set mapFile $filename
	set fdir [file dirname $filename]
        catch {cd $fdir}
    } else {
        tk_messageBox -title $title -icon error -type ok \
		-message "File '$filename' is not found."
	return
    }

    if {[string length [setupMap .]] > 0} {
        update
        drawImages .
        setMsg "mapfile [file tail $mapFile] loaded, select layers to view\
             and Refresh"
    } else {
        setMsg "mapfile [file tail $mapFile] could not be loaded"
    }
}

proc setMsg {txt} {
    .msg configure -text $txt
    update idletasks
}

proc getInterps {} {
    global tcl_platform

    if {"$tcl_platform(platform)" == "windows"} {
	update
        set all_interps [dde services TclEval {}]
        set interps ""
        foreach intrp $all_interps {
            lappend interps [lindex $intrp 1]
        }
    } else {
        set interps [winfo interps]
    }
    return $interps
}

proc startMSExe {exe_name} {
    global mapobj mapfile startupdir tcl_platform title
    if {[lsearch [getInterps] $exe_name] == -1} {
        if {"$tcl_platform(platform)" == "windows"} {
            set start_exe [list exec start [file join $startupdir $exe_name] &]
            set error_msg "Cannot 'send' to $exe_name via DDE emulation."
        } else {
            set start_exe [list exec [file join $startupdir $exe_name] &]
            set error_msg "Cannot 'send' to $exe_name. Check your xhost\
                settings, xhost should report\
                \n\"access control enabled, only authorized clients can\
                 connect\" "
        }
        if {[catch $start_exe result]} {
            tk_messageBox -icon error -type ok \
                -title $title \
                -message "Cannot start $exe_name\n\n$result"
            return 0
        } else {
            setMsg "please wait, starting $exe_name"
            # wait for Tk program to startup
            set i 20
            while {$i >= 0 && [lsearch [getInterps] $exe_name] == -1} {
                update
                after 1000
                incr i -1
                update
            }
            if {$i <= 0 && [lsearch [getInterps] $exe_name] == -1} {
                tk_messageBox -icon error -type ok \
                     -title $title \
                     -message "$error_msg\n\n$result"
                setMsg "could not contact $exe_name"
                return 0
            }
            after 1000
        }
    }
    update
    if {[catch {send $exe_name update} result]} {
        tk_messageBox -icon error -type ok -title $title \
                     -message "$error_msg\n\n$result"
	return 0
    }

    return 1
}


proc startMSed {} {
    global mapFile startupdir title
    set exe_name msedit.tcl

    if {! [startMSExe $exe_name]} {
	return
    }
    if {[string length $mapFile] == 0} {
        return
    }
    update
    if {[catch {send $exe_name remoteOpenFile $mapFile} result]} {
        tk_messageBox -icon error -type ok -title $title \
                     -message "Cannot 'send' to $exe_name:\n\n$result"
    }
    setMsg "ready"
}




############################################################################
# 'send' emulation for windows
#
# written by George A. Howlett, 
# scarfed from a comp.lang.tcl posting by George.
#
#

# --------------------------------------------------------------------------
#
# SendInit --
#	
#	Creates a "send" proc to replace the former Tk send command.
#	Uses DDE services to simulate the transfer.  This must be
#	called before any drag&drop targets are registered. Otherwise
#	they will pick up the wrong application name.
#
#	The first trick is to determine a unique application name. This
#	is what other applications will use to send to us.  Tk used to
#	do this for us.  
#	
#	Note that we can generate the same name for two different Tk 
#	applications.  This can happen if two Tk applications picking
#	names at exactly the same time.   [In the future, we should 
#	probably generate a name based upon a global system value, such
#	as the handle of the main window "."]   The proc "SendVerify"
#	below will verify that you have only one DDE server registered
#	with this application's name.
#	
# Arguments:
#	myInterp	Sets the application name explicitly to this 
#			string. If the argument isn't given, or is the
#			empty string, then the routine picks a name for
#			us.
#
# Results:
#	Returns the name of the application.
#
# Side Effects:
#	Sets the name of our application.  You can call "tk appname" to
#	get the name.  A DDE topic using the same name is also created.
#	A send proc is also automatically created.  Be careful that you
#	don't overwrite an existing send command.
#
# --------------------------------------------------------------------------

proc SendInit { {myInterp ""} } {

    # Load the DDE package.
    package require dde

    if { $myInterp == "" } {

	# Pick a unique application name, replicating what Tk used to do.
	# This is what other applications will use to "send" to us. We'll
	# use DDE topics to represent interpreters.

	set appName [tk appname]
	set count 0
	set suffix {}
	
	# Keep generating interpreter names by suffix-ing the original
	# application name with " #number".  Sooner of later we'll find
	# one that's not currently use. 

	while { 1 } {
	    set myInterp "${appName}${suffix}"
	    set myServer [list TclEval $myInterp]
	    if { [lsearch [dde services TclEval {}] $myServer] < 0 } {
		break
	    }
	    incr count
	    set suffix " \#$count"
	}
    }
    tk appname $myInterp
    dde servername $myInterp
    proc send { interp args } {
	dde eval $interp $args
    }
    return $myInterp
}


# --------------------------------------------------------------------------
#
# SendVerify --
#	
#	Verifies that application name picked is uniquely registered
#	as a DDE server.  This checks that two Tk applications don't
#	accidently use the same name.
#
# Arguments:
#	None		Used the current application name.
#
# Results:
#	Generates an error if either a server can't be found or more
#	than one server is registered.
#
# --------------------------------------------------------------------------

proc SendVerify {} {
    # Load the DDE package.
    package require dde

    set count 0
    set appName [tk appname]
    foreach server [dde services TclEval {}] {
	set topic [lindex $server 1]
	if { [string compare $topic $appName] == 0 } {
	    incr count
	}
    }
    if {$count == 0} {
	error "Service not found: wrong name registered???"
    } 
    if { $count > 1 } {
	error "Duplicate names found for \"[tk appname]\""
    } 
}

#
###########################################################################


# main code


if {"$tcl_platform(platform)" == "windows"} {
    set scriptname [file tail $scriptname]
    SendInit $scriptname
    if {[catch {SendVerify} reason]} {
        tk_messageBox -icon error -type warning \
                -title $title \
                -message "Cannot initialize \"send\" emulation using DDE.\n\
                    \n$reason\n\
                    \nInteraction with other MapServer Workbench tools\
                    may not function properly."
    }
}

# Lay out UI
setupWindow .
setMsg "ready"

