#!/bin/sh
# The next line is executed by /bin/sh, but not Tcl \
    exec /usr/local/bin/wish8.3 $0 ${1+"$@"}

# if you are installing by hand,
# replace /usr/local/bin/wish8.3 
# on line 3 with the full path to your wish executable, verison 8.3 or higher.

global startupdir
set startupdir ""
set scriptname [info script]
# on unix, chase symbolic links until we get to the real program
if {"$tcl_platform(platform)" == "unix"} {
    while {[catch {set scriptname [file readlink $scriptname]}] == 0} { }
}
set startupdir [file dirname $scriptname]
if {"$startupdir" == "."} {
     set startupdir [pwd]
}
lappend auto_path $startupdir [file dirname $startupdir]

package require mkWidgets
package require Mapscript
package require msTclObjs

option add *Dialog.msg.wrapLength 700
option add *Dialog.msg.font {helvetica 12 bold}

if {[info exists env(TEMP)]} {
  set tmpdir $env(TEMP)
} else {
  # Urgh
  set tmpdir /tmp
  catch {file mkdir /tmp}
}

global title
set title "MapServer Shape Explorer"

global shapefile	
set shapefile ""

global mapfile	
set mapfile ""

global shapetfileypes
set shapefiletypes {
    {{Shape Files}     {.shp}        }
    {{All Files}        *            }
}

global types map_types
array set types {
   1 point
   3 arc
   5 polygon
   8 multipoint
}
array set map_types {
   1 point
   3 line
   5 polygon
   8 point
}

global currec maxrec
set currec 0
set maxrec 0

global dbf_names dbf_types
set dbf_names ""
set dbf_types ""

global hdbf
set hdbf ""

global scrl
set scrl ""
global grid
set grid ""

global cleanup
set cleanup ""

proc mkMain {} {
    global scrl grid title
   
    wm title . $title
    wm protocol . WM_DELETE_WINDOW doExit
    wm geom . 750x500

    menu .menuBar -tearoff 0

    .menuBar add cascade -menu .menuBar.file -label "File" -underline 0
    .menuBar add cascade -menu .menuBar.tool -label "Tools" -underline 0
    .menuBar add cascade -menu .menuBar.help -label "Help" -underline 0

    menu .menuBar.file -tearoff 0
    .menuBar.file add command -label "Open" -underline 0 -command openShape \
	-accelerator Ctrl+O
    bind all <Control-KeyPress-o> +openShape
    .menuBar.file add separator
    .menuBar.file add command -label "Quit" -underline 0 -command doExit \
	-accelerator Ctrl+Q
    bind all <Control-KeyPress-q> +doExit

    menu .menuBar.tool -tearoff 0
    .menuBar.tool add command -label "Clear Records" \
	-command clearGrid -underline 0 -state disabled -accelerator Ctrl+C
    bind all <Control-KeyPress-c> +clearGrid

    .menuBar.tool add separator
    .menuBar.tool add cascade -label "View shape in MapServer Explorer" \
	-menu .menuBar.tool.view -underline 0 -state disabled

    menu .menuBar.tool.view -tearoff 0
    .menuBar.tool.view add command -label "Label features with: No labels" \
	-command {viewShapeFile ""}
    .menuBar.tool.view add separator

    menu .menuBar.help -tearoff 0
    .menuBar.help add command -label "About" -command loadHelp -underline 0


    . configure -menu .menuBar
    
    pane .top -resize both -orientation vertical -width 3
    frame .top.left  -width 300
    frame .top.right  -width 600
    .top pack first  .top.left  -fill both -expand 1 
    .top pack second .top.right -fill both -expand 1

    set scrl .top.left.scrl
    scrollbox $scrl -scrollbar auto
    pack .top.left.scrl -expand 1 -fill both -padx 2 -pady 2

    frame .top.right.bar -borderwidth 4 -relief sunken 
    button .top.right.bar.next -text "Show next 20 records" -command showDBF
    label .top.right.bar.l1 -text " starting at:"
    entry .top.right.bar.en -textvariable currec  -width 8
    pack .top.right.bar.next .top.right.bar.l1 .top.right.bar.en -side left
    pack .top.right.bar -side top -fill x -pady 10 -padx 5

    set grid .top.right.grid
    pack [gridcontrol $grid] -side top -expand 1 -fill both


    pack .top       -side top -fill both -expand 1
    
    label .msg -text ready -anchor w
    pack .msg -side bottom -fill x -anchor w

    
    frame .bottom 
    pack  .bottom -side top -fill x 

    update 
}


proc loadHelp {} {
    global title
    tk_messageBox -title $title -message \
     "$title\nTom Poindexter\ntpoindex@nyx.net\
      \nCopyright 2000\n\n\
    Notes:\n\
    Records begin at index 0, but are displayed starting with 1.\n\n\
    "
}


proc clearGrid {} {
    global grid currec dbf_names
    
    if {[llength $dbf_names] > 0} {
        catch {destroy $grid}
        pack [gridcontrol $grid] -side top -expand 1 -fill both
	populateGrid
        set currec 0
    }
}


proc setMsg {txt} {
    .msg configure -text $txt
    update idletasks
}

proc doExit {} {
    global cleanup
    if {[string length $cleanup]} {
	catch {file delete $cleanup}
    }
    exit
}


proc openShape {} {
    global shapefiletypes shapefile

    set shapefile [tk_getOpenFile -filetypes $shapefiletypes -parent .]
    if {$shapefile == {}} {
	return
    }
    catch {cd [file dirname $shapefile]}
    openShapeFile
}

proc openShapeFile {} {      
    global title
    global shapefile scrl grid types currec maxrec hdbf dbf_names dbf_types
    set currec 0
    set maxrec 0
    set hdbf ""
    $scrl delete 0 end
    
    catch {destroy $grid}
    pack [gridcontrol $grid] -side top -expand 1 -fill both
    set dbf_names ""

    catch {rename shp {}}
    set rc ""
    set rc [mapscript::shapefileObj shp $shapefile -1]
    if {"$rc" == "NULL"} {
	tk_messageBox -title $title -message "Error opening shapefile: \
                 [file tail $shapefile]\n\n\
        "
	set shapefile ""
	return
    }

    $scrl insert end " Shapefile: [file tail $shapefile]"
    $scrl insert end "   type: $types([shp cget -type])"
    $scrl insert end "   # of features: [shp cget -numshapes]"
    $scrl insert end " bounds:"
    set rectPtr [shp cget -bounds]
    $scrl insert end  "   minx [mapscript::rectObjRef $rectPtr cget -minx]"
    $scrl insert end  "   miny [mapscript::rectObjRef $rectPtr cget -miny]"
    $scrl insert end  "   maxx [mapscript::rectObjRef $rectPtr cget -maxx]"
    $scrl insert end  "   maxy [mapscript::rectObjRef $rectPtr cget -maxy]"
    
    setMsg "opened shapefile $shapefile"
    
    set maxrec [shp cget -numshapes]
    set currec 0
    catch {rename $rectPtr {}}

    set hdbf [shp cget -hDBF]
    if {"$hdbf" == "NULL"} {
	tk_messageBox -title $title -type error -message "Shapefile has missing\
               or damaged .dbf file"
        return
    }

    set dbf_names [mapscript::getDBFNames $hdbf]
    set dbf_types [mapscript::getDBFTypes $hdbf]

    populateGrid

    .menuBar.tool entryconfigure 0 -state normal
    .menuBar.tool entryconfigure 2 -state normal
    .menuBar.tool.view delete 0 999
    .menuBar.tool.view add command -label "Label features with: No labels" \
	-command {viewShapeFile ""}
    .menuBar.tool.view add separator

    $scrl insert end  " attributes:"
    set i 0
    foreach n $dbf_names {
        .menuBar.tool.view add command -label "      $n" \
		-command [list viewShapeFile $n]
	set ctyp [lindex [lindex $dbf_types $i] 0]
	set clen [lindex [lindex $dbf_types $i] 1]
	$scrl insert end  "   $n  ($ctyp $clen)"
	incr i
    }
    
}

proc populateGrid {} {
    global dbf_names dbf_types grid
    set i 0
    foreach n $dbf_names {
        set ctyp [lindex [lindex $dbf_types $i] 0]
        set clen [lindex [lindex $dbf_types $i] 1]
        switch $ctyp {
	    double  {set just numeric}
	    integer {set just right}
	    string  -
	    default {set just left}
        }
	set wid [expr {$clen > [string length $n] ? $clen : [string length $n]}]
	incr wid 2
	set wid [expr {$wid * 10}]
	$grid column insert $n end -text $n -width $wid -align $just
	incr i
    }
}


proc showDBF {} {
    global grid types currec maxrec hdbf dbf_names dbf_types
    if {"$hdbf" == "NULL" || [string length $hdbf] == 0} {
	set currec 0
	return
    }
    if {![string is integer -strict $currec]} {
	set currec 0
    }
    if {$currec < 0} {
	set currec 0
    }
    if {$currec >= $maxrec} {
	set currec [expr {$maxrec - 1}]
    }
    for {set i 0} {$i < 20 && $currec < $maxrec} {incr i; incr currec} {
	$grid insert end [mapscript::getDBFValues $hdbf $currec]
    }
}

proc remoteOpenFile {filename} {
    global shapefile
    wm deiconify .
    raise .

    set shapefile $filename
    catch {cd [file dirname $shapefile]}
    openShapeFile
}

proc viewShapeFile {label_item} {
    global mapfile shapefile tmpdir dbf_names cleanup map_types

    if {! [string length $shapefile]} {
	setMsg "no shapefile is open"
        return
    }

    set name [file rootname [file tail $shapefile]]
    
    # construct minimal mapfile with layer
    set mapobj [mstclobj::mapTclObj #auto]
    $mapobj setAttr name       $name
    $mapobj setAttr imagecolor "255 255 255"
    $mapobj setAttr status     on
    $mapobj setAttr size       "600 400"
    $mapobj setAttr shapepath  [pwd]

    set rectPtr [shp cget -bounds]
    foreach v {minx miny maxx maxy} {
	set $v [mapscript::rectObjRef $rectPtr cget -$v]
    }
    $mapobj setAttr extent "$minx $miny $maxx $maxy"

    set legobj [mstclobj::legendTclObj #auto $mapobj]
    $legobj setAttr imagecolor "134 72 134"
    $legobj setAttr keysize    "18 12"
    $legobj setAttr status     on
    set labobj [mstclobj::labelTclObj #auto $legobj]
    $labobj setAttr type       bitmap
    $labobj setAttr size       medium
    $labobj setAttr color      "200 40 89"

    set layerobj [mstclobj::layerTclObj #auto $mapobj]
    $layerobj setAttr name $name
    $layerobj setAttr classitem [lindex $dbf_names 0]
    $layerobj setAttr data      [file tail $shapefile]
    $layerobj setAttr status    on
    $layerobj setAttr type      $map_types([shp cget -type])
    if {[string length $label_item]} {
        $layerobj setAttr labelitem $label_item
    }
 
    set classobj [mstclobj::classTclObj #auto $layerobj]
    $classobj setAttr color           "89 132 46"
    $classobj setAttr outlinecolor    "0 0 0"
    $classobj setAttr backgroundcolor "128 178 12"
    if {[string length $label_item]} {
	set labelobj [mstclobj::labelTclObj #auto $classobj]
	$labelobj setAttr color           "255 255 255"
	$labelobj setAttr backgroundcolor "89 132 46"
	$labelobj setAttr size            tiny
    }

    if {[string length $cleanup]} {
	catch {file delete $cleanup}
        set cleanup ""
    }
    set mapfile $tmpdir/[$mapobj getAttr name][pid].map
    
    $mapobj dumpToFile $mapfile
    startMSex
    
    ::obstcl::delete $mapobj
    set cleanup $mapfile
}

proc getInterps {} {
    global tcl_platform

    if {"$tcl_platform(platform)" == "windows"} {
	update
	set all_interps [dde services TclEval {}]
	set interps ""
	foreach intrp $all_interps {
	    lappend interps [lindex $intrp 1]
	}
    } else {
	set interps [winfo interps]
    }
    return $interps
}

proc startMSExe {exe_name} {
    global mapobj mapfile startupdir tcl_platform title
    if {[lsearch [getInterps] $exe_name] == -1} {
	if {"$tcl_platform(platform)" == "windows"} {
	    set start_exe [list exec start [file join $startupdir $exe_name] &]
	    set error_msg "Cannot 'send' to $exe_name via DDE emulation."
	} else {
	    set start_exe [list exec [file join $startupdir $exe_name] &]
	    set error_msg "Cannot 'send' to $exe_name. Check your xhost\
                settings, xhost should report\
                \n\"access control enabled, only authorized clients can\
                 connect\" "
	}
        if {[catch $start_exe result]} {
            tk_messageBox -icon error -type ok \
		-title $title \
                -message "Cannot start $exe_name\n\n$result"
            return 0
        } else {
            setMsg "please wait, starting $exe_name"
            # wait for Tk program to startup
            set i 20
            while {$i >= 0 && [lsearch [getInterps] $exe_name] == -1} {
		update
                after 1000
                incr i -1
		update
            }
            if {$i <= 0 && [lsearch [getInterps] $exe_name] == -1} {
                tk_messageBox -icon error -type ok \
                     -title $title \
                     -message "$error_msg\n\n$result"
                setMsg "could not contact $exe_name"
                return 0
            }
            after 1000
        }
    }
    update
    if {[catch {send $exe_name update} result]} {
        tk_messageBox -icon error -type ok -title $title \
                     -message "$error_msg\n\n$result"
        return 0
    }

    return 1
}

proc startMSex {} {
    global mapobj mapfile startupdir tcl_platform title
    set exe_name msexplorer.tcl

    if {! [startMSExe $exe_name]} {
	return
    }
    set mfile  $mapfile
    update
    if {[catch {send $exe_name remoteOpenFile $mfile} result]} {
        tk_messageBox -icon error -type ok -title $title \
                     -message "Cannot 'send' to $exe_name:\n\n\$result"
    }
}



############################################################################
# 'send' emulation for windows
#
# written by George A. Howlett, 
# scarfed from a comp.lang.tcl posting by George.
#
#

# --------------------------------------------------------------------------
#
# SendInit --
#	
#	Creates a "send" proc to replace the former Tk send command.
#	Uses DDE services to simulate the transfer.  This must be
#	called before any drag&drop targets are registered. Otherwise
#	they will pick up the wrong application name.
#
#	The first trick is to determine a unique application name. This
#	is what other applications will use to send to us.  Tk used to
#	do this for us.  
#	
#	Note that we can generate the same name for two different Tk 
#	applications.  This can happen if two Tk applications picking
#	names at exactly the same time.   [In the future, we should 
#	probably generate a name based upon a global system value, such
#	as the handle of the main window "."]   The proc "SendVerify"
#	below will verify that you have only one DDE server registered
#	with this application's name.
#	
# Arguments:
#	myInterp	Sets the application name explicitly to this 
#			string. If the argument isn't given, or is the
#			empty string, then the routine picks a name for
#			us.
#
# Results:
#	Returns the name of the application.
#
# Side Effects:
#	Sets the name of our application.  You can call "tk appname" to
#	get the name.  A DDE topic using the same name is also created.
#	A send proc is also automatically created.  Be careful that you
#	don't overwrite an existing send command.
#
# --------------------------------------------------------------------------

proc SendInit { {myInterp ""} } {

    # Load the DDE package.
    package require dde

    if { $myInterp == "" } {

	# Pick a unique application name, replicating what Tk used to do.
	# This is what other applications will use to "send" to us. We'll
	# use DDE topics to represent interpreters.

	set appName [tk appname]
	set count 0
	set suffix {}
	
	# Keep generating interpreter names by suffix-ing the original
	# application name with " #number".  Sooner of later we'll find
	# one that's not currently use. 

	while { 1 } {
	    set myInterp "${appName}${suffix}"
	    set myServer [list TclEval $myInterp]
	    if { [lsearch [dde services TclEval {}] $myServer] < 0 } {
		break
	    }
	    incr count
	    set suffix " \#$count"
	}
    }
    tk appname $myInterp
    dde servername $myInterp
    proc send { interp args } {
	dde eval $interp $args
    }
    return $myInterp
}


# --------------------------------------------------------------------------
#
# SendVerify --
#	
#	Verifies that application name picked is uniquely registered
#	as a DDE server.  This checks that two Tk applications don't
#	accidently use the same name.
#
# Arguments:
#	None		Used the current application name.
#
# Results:
#	Generates an error if either a server can't be found or more
#	than one server is registered.
#
# --------------------------------------------------------------------------

proc SendVerify {} {
    # Load the DDE package.
    package require dde

    set count 0
    set appName [tk appname]
    foreach server [dde services TclEval {}] {
	set topic [lindex $server 1]
	if { [string compare $topic $appName] == 0 } {
	    incr count
	}
    }
    if {$count == 0} {
	error "Service not found: wrong name registered???"
    } 
    if { $count > 1 } {
	error "Duplicate names found for \"[tk appname]\""
    } 
}

#
###########################################################################


# main code


if {"$tcl_platform(platform)" == "windows"} {
    set scriptname [file tail $scriptname]
    SendInit $scriptname
    if {[catch {SendVerify} reason]} {
        tk_messageBox -icon warning -type ok \
                -title $title \
                -message "Cannot initialize \"send\" emulation using DDE:\n\
                    \n$reason\n\
                    \nInteraction with other MapServer Workbench tools\
                    may not function properly."
    }
}



mkMain

setMsg "ready"

